@@ -1,10 +1,10 @@
 /* reading patches */
 
-/* $Id: pch.c,v 1.26 1997/07/21 17:59:46 eggert Exp $ */
+/* $Id: pch.c,v 1.27 1998/03/15 14:44:47 eggert Exp $ */
 
 /*
 Copyright 1986, 1987, 1988 Larry Wall
-Copyright 1990, 1991, 1992, 1993, 1997 Free Software Foundation, Inc.
+Copyright 1990, 1991, 1992, 1993, 1997, 1998 Free Software Foundation, Inc.
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -26,6 +26,7 @@ If not, write to the Free Software Foundation,
 #include <common.h>
 #include <backupfile.h>
 #include <inp.h>
+#include <quotearg.h>
 #include <util.h>
 #undef XTERN
 #define XTERN
@@ -55,6 +56,7 @@ static size_t *p_len;			/* line length including \n if any */
 static char *p_Char;			/* +, -, and ! */
 static LINENUM hunkmax = INITHUNKMAX;	/* size of above arrays */
 static int p_indent;			/* indent to patch */
+static int p_strip_trailing_cr;		/* nonzero if stripping trailing \r */
 static file_offset p_base;		/* where to intuit this time */
 static LINENUM p_bline;			/* line # of p_base */
 static file_offset p_start;		/* where intuit found a patch */
@@ -65,10 +67,11 @@ static LINENUM p_bfake = -1;		/* beg of faked up lines */
 
 enum nametype { OLD, NEW, INDEX, NONE };
 
+static char *scan_linenum PARAMS ((char *, LINENUM *));
 static enum diff intuit_diff_type PARAMS ((void));
 static enum nametype best_name PARAMS ((char * const *, int const *));
 static int prefix_components PARAMS ((char *, int));
-static size_t pget_line PARAMS ((int, int));
+static size_t pget_line PARAMS ((int, int, int));
 static size_t get_line PARAMS ((void));
 static bool incomplete_line PARAMS ((void));
 static bool grow_hunkmax PARAMS ((void));
@@ -88,6 +91,7 @@ re_patch()
     p_end = -1;
     p_max = 0;
     p_indent = 0;
+    p_strip_trailing_cr = 0;
 }
 
 /* Open the patch file at the beginning of time. */
@@ -119,9 +123,14 @@ open_patch_file(filename)
 	else
 	  {
 	    size_t charsread;
-	    pfp = fopen (TMPPATNAME, "w+b");
+	    int exclusive = TMPPATNAME_needs_removal ? 0 : O_EXCL;
+	    TMPPATNAME_needs_removal = 1;
+	    pfp = fdopen (create_file (TMPPATNAME,
+				       O_RDWR | O_BINARY | exclusive,
+				       (mode_t) 0),
+			  "w+b");
 	    if (!pfp)
-	      pfatal ("can't create `%s'", TMPPATNAME);
+	      pfatal ("Can't open stream for file %s", quotearg (TMPPATNAME));
 	    for (st.st_size = 0;
 		 (charsread = fread (buf, 1, bufsize, stdin)) != 0;
 		 st.st_size += charsread)
@@ -138,7 +147,7 @@ open_patch_file(filename)
       {
 	pfp = fopen (filename, binary_transput ? "rb" : "r");
 	if (!pfp)
-	  pfatal ("can't open patch file `%s'", filename);
+	  pfatal ("Can't open patch file %s", quotearg (filename));
 	if (fstat (fileno (pfp), &st) != 0)
 	  pfatal ("fstat");
       }
@@ -223,11 +232,14 @@ there_is_another_patch()
       {
 	if (p_indent)
 	  say ("(Patch is indented %d space%s.)\n", p_indent, p_indent==1?"":"s");
+	if (p_strip_trailing_cr)
+	  say ("(Stripping trailing CRs from patch.)\n");
 	if (! inname)
 	  {
-	    say ("can't find file to patch at input line %ld\n",
-		 p_sline);
-	    say (strippath == INT_MAX
+	    char numbuf[LINENUM_LENGTH_BOUND + 1];
+	    say ("can't find file to patch at input line %s\n",
+		 format_linenum (numbuf, p_sline));
+	    say (strippath == -1
 	         ? "Perhaps you should have used the -p or --strip option?\n"
 		 : "Perhaps you used the wrong -p or --strip option?\n");
 	  }
@@ -274,16 +286,10 @@ there_is_another_patch()
 static enum diff
 intuit_diff_type()
 {
-    register char *s;
-    register char *t;
-    register int indent;
     register file_offset this_line = 0;
-    register file_offset previous_line;
     register file_offset first_command_line = -1;
     LINENUM fcl_line = 0; /* Pacify `gcc -W'.  */
-    register bool last_line_was_command = FALSE;
     register bool this_is_a_command = FALSE;
-    register bool stars_last_line = FALSE;
     register bool stars_this_line = FALSE;
     enum nametype i;
     char *name[3];
@@ -302,12 +308,20 @@ intuit_diff_type()
     Fseek (pfp, p_base, SEEK_SET);
     p_input_line = p_bline - 1;
     for (;;) {
-	previous_line = this_line;
-	last_line_was_command = this_is_a_command;
-	stars_last_line = stars_this_line;
+	register char *s;
+	register char *t;
+	register file_offset previous_line = this_line;
+	register bool last_line_was_command = this_is_a_command;
+	register bool stars_last_line = stars_this_line;
+	register int indent = 0;
+	int strip_trailing_cr;
+	size_t chars_read;
+
 	this_line = file_tell (pfp);
-	indent = 0;
-	if (! pget_line (0, 0)) {
+	chars_read = pget_line (0, 0, 0);
+	if (chars_read == (size_t) -1)
+	  memory_fatal ();
+	if (! chars_read) {
 	    if (first_command_line >= 0) {
 					/* nothing but deletes!? */
 		p_start = first_command_line;
@@ -321,6 +335,7 @@ intuit_diff_type()
 		return NO_DIFF;
 	    }
 	}
+	strip_trailing_cr = 2 <= chars_read && buf[chars_read - 2] == '\r';
 	for (s = buf; *s == ' ' || *s == '\t' || *s == 'X'; s++) {
 	    if (*s == '\t')
 		indent = (indent + 8) & ~7;
@@ -335,6 +350,7 @@ intuit_diff_type()
 	    first_command_line = this_line;
 	    fcl_line = p_input_line;
 	    p_indent = indent;		/* assume this for now */
+	    p_strip_trailing_cr = strip_trailing_cr;
 	}
 	if (!stars_last_line && strnEQ(s, "*** ", 4))
 	    name[OLD] = fetchname (s+4, strippath, &p_timestamp[OLD]);
@@ -376,6 +392,7 @@ intuit_diff_type()
 	  first_command_line >= 0 &&
 	  strEQ(s, ".\n") ) {
 	    p_indent = indent;
+	    p_strip_trailing_cr = strip_trailing_cr;
 	    p_start = first_command_line;
 	    p_sline = fcl_line;
 	    retval = ED_DIFF;
@@ -393,22 +410,26 @@ intuit_diff_type()
 	    name[NEW] = t;
 
 	    s += 4;
-	    if (! atol (s))
+	    if (s[0] == '0' && !ISDIGIT (s[1]))
 	      p_says_nonexistent[OLD] = 1 + ! p_timestamp[OLD];
 	    while (*s != ' ' && *s != '\n')
 	      s++;
 	    while (*s == ' ')
 	      s++;
-	    if (! atol (s))
+	    if (s[0] == '+' && s[1] == '0' && !ISDIGIT (s[2]))
 	      p_says_nonexistent[NEW] = 1 + ! p_timestamp[NEW];
 	    p_indent = indent;
+	    p_strip_trailing_cr = strip_trailing_cr;
 	    p_start = this_line;
 	    p_sline = p_input_line;
 	    retval = UNI_DIFF;
 	    if (! ((name[OLD] || ! p_timestamp[OLD])
 		   && (name[NEW] || ! p_timestamp[NEW])))
-	      say ("missing header for unified diff at line %ld of patch\n",
-		   p_sline);
+	      {
+		char numbuf[LINENUM_LENGTH_BOUND + 1];
+		say ("missing header for unified diff at line %s of patch\n",
+		     format_linenum (numbuf, p_sline));
+	      }
 	    goto scan_exit;
 	}
 	stars_this_line = strnEQ(s, "********", 8);
@@ -417,13 +438,14 @@ intuit_diff_type()
 	     || diff_type == NEW_CONTEXT_DIFF)
 	    && stars_last_line && strnEQ (s, "*** ", 4)) {
 	    s += 4;
-	    if (! atol (s))
+	    if (s[0] == '0' && !ISDIGIT (s[1]))
 	      p_says_nonexistent[OLD] = 1 + ! p_timestamp[OLD];
 	    /* if this is a new context diff the character just before */
 	    /* the newline is a '*'. */
 	    while (*s != '\n')
 		s++;
 	    p_indent = indent;
+	    p_strip_trailing_cr = strip_trailing_cr;
 	    p_start = previous_line;
 	    p_sline = p_input_line - 1;
 	    retval = (*(s-1) == '*' ? NEW_CONTEXT_DIFF : CONTEXT_DIFF);
@@ -443,8 +465,11 @@ intuit_diff_type()
 
 	    if (! ((name[OLD] || ! p_timestamp[OLD])
 		   && (name[NEW] || ! p_timestamp[NEW])))
-	      say ("missing header for context diff at line %ld of patch\n",
-		   p_sline);
+	      {
+		char numbuf[LINENUM_LENGTH_BOUND + 1];
+		say ("missing header for context diff at line %s of patch\n",
+		     format_linenum (numbuf, p_sline));
+	      }
 	    goto scan_exit;
 	}
 	if ((diff_type == NO_DIFF || diff_type == NORMAL_DIFF) &&
@@ -453,6 +478,7 @@ intuit_diff_type()
 	    p_start = previous_line;
 	    p_sline = p_input_line - 1;
 	    p_indent = indent;
+	    p_strip_trailing_cr = strip_trailing_cr;
 	    retval = NORMAL_DIFF;
 	    goto scan_exit;
 	}
@@ -545,7 +571,8 @@ intuit_diff_type()
 				version_controlled[i] = 0;
 
 			      free (getbuf);
-			      free (diffbuf);
+			      if (diffbuf)
+				free (diffbuf);
 
 			      if (! stat_errno[i])
 				break;
@@ -560,12 +587,12 @@ intuit_diff_type()
 	      {
 		assert (i0 != NONE);
 		if (ok_to_reverse
-		    ("The next patch%s would %s the file `%s',\nwhich %s!",
+		    ("The next patch%s would %s the file %s,\nwhich %s!",
 		     reverse ? ", when reversed," : "",
 		     (i == NONE ? "delete"
 		      : st[i].st_size == 0 ? "empty out"
 		      : "create"),
-		     name[i == NONE || st[i].st_size == 0 ? i0 : i],
+		     quotearg (name[i == NONE || st[i].st_size == 0 ? i0 : i]),
 		     (i == NONE ? "does not exist"
 		      : st[i].st_size == 0 ? "is already empty"
 		      : "already exists")))
@@ -657,9 +684,9 @@ best_name (name, ignore)
   int components[3];
   int components_min = INT_MAX;
   size_t basename_len[3];
-  size_t basename_len_min = (size_t) -1;
+  size_t basename_len_min = SIZE_MAX;
   size_t len[3];
-  size_t len_min = (size_t) -1;
+  size_t len_min = SIZE_MAX;
 
   for (i = OLD;  i <= INDEX;  i++)
     if (name[i] && !ignore[i])
@@ -744,10 +771,43 @@ LINENUM file_line;
 static void
 malformed ()
 {
-    fatal ("malformed patch at line %ld: %s", p_input_line, buf);
+    char numbuf[LINENUM_LENGTH_BOUND + 1];
+    fatal ("malformed patch at line %s: %s",
+	   format_linenum (numbuf, p_input_line), buf);
 		/* about as informative as "Syntax error" in C */
 }
 
+/* Parse a line number from a string.
+   Return the address of the first char after the number.  */
+static char *
+scan_linenum (s0, linenum)
+     char *s0;
+     LINENUM *linenum;
+{
+  char *s;
+  LINENUM n = 0;
+  int overflow = 0;
+  char numbuf[LINENUM_LENGTH_BOUND + 1];
+
+  for (s = s0;  ISDIGIT (*s);  s++)
+    {
+      LINENUM new_n = 10 * n + (*s - '0');
+      overflow |= new_n / 10 != n;
+      n = new_n;
+    }
+  
+  if (s == s0)
+    fatal ("missing line number at line %s: %s",
+	   format_linenum (numbuf, p_input_line), buf);
+  
+  if (overflow)
+    fatal ("line number %.*s is too large at line %s: %s",
+	   (int) (s - s0), s0, format_linenum (numbuf, p_input_line), buf);
+
+  *linenum = n;
+  return s;
+}
+
 /* 1 if there is more of the current diff listing to process;
    0 if not; -1 if ran out of memory. */
 
@@ -759,6 +819,10 @@ another_hunk (difftype, rev)
     register char *s;
     register LINENUM context = 0;
     register size_t chars_read;
+    char numbuf0[LINENUM_LENGTH_BOUND + 1];
+    char numbuf1[LINENUM_LENGTH_BOUND + 1];
+    char numbuf2[LINENUM_LENGTH_BOUND + 1];
+    char numbuf3[LINENUM_LENGTH_BOUND + 1];
 
     while (p_end >= 0) {
 	if (p_end == p_efake)
@@ -821,8 +885,9 @@ another_hunk (difftype, rev)
 	    }
 	    p_end++;
 	    if (p_end == hunkmax)
-		fatal ("unterminated hunk starting at line %ld; giving up at line %ld: %s",
-		       pch_hunk_beg (), p_input_line, buf);
+	      fatal ("unterminated hunk starting at line %s; giving up at line %s: %s",
+		     format_linenum (numbuf0, pch_hunk_beg ()),
+		     format_linenum (numbuf1, p_input_line), buf);
 	    assert(p_end < hunkmax);
 	    p_Char[p_end] = *buf;
 	    p_len[p_end] = 0;
@@ -835,16 +900,16 @@ another_hunk (difftype, rev)
 			goto hunk_done;
 		    }
 		    else
-			fatal ("unexpected end of hunk at line %ld",
-			    p_input_line);
+		      fatal ("unexpected end of hunk at line %s",
+			     format_linenum (numbuf0, p_input_line));
 		}
 		if (p_end != 0) {
 		    if (repl_beginning && repl_could_be_missing) {
 			repl_missing = TRUE;
 			goto hunk_done;
 		    }
-		    fatal ("unexpected `***' at line %ld: %s",
-			   p_input_line, buf);
+		    fatal ("unexpected `***' at line %s: %s",
+			   format_linenum (numbuf0, p_input_line), buf);
 		}
 		context = 0;
 		p_len[p_end] = strlen (buf);
@@ -854,19 +919,14 @@ another_hunk (difftype, rev)
 		}
 		for (s = buf;  *s && !ISDIGIT (*s);  s++)
 		  continue;
-		if (!*s)
-		    malformed ();
 		if (strnEQ(s,"0,0",3))
 		    remove_prefix (s, 2);
-		p_first = (LINENUM) atol(s);
-		while (ISDIGIT (*s))
-		  s++;
+		s = scan_linenum (s, &p_first);
 		if (*s == ',') {
 		    while (*s && !ISDIGIT (*s))
 		      s++;
-		    if (!*s)
-			malformed ();
-		    p_ptrn_lines = ((LINENUM)atol(s)) - p_first + 1;
+		    scan_linenum (s, &p_ptrn_lines);
+		    p_ptrn_lines += 1 - p_first;
 		}
 		else if (p_first)
 		    p_ptrn_lines = 1;
@@ -901,14 +961,17 @@ another_hunk (difftype, rev)
 			fillcnt = p_ptrn_lines;
 		      }
 		    else if (! repl_beginning)
-		      fatal ("%s `---' at line %ld; check line numbers at line %ld",
+		      fatal ("%s `---' at line %s; check line numbers at line %s",
 			     (p_end <= p_ptrn_lines
 			      ? "Premature"
 			      : "Overdue"),
-			     p_input_line, p_hunk_beg);
+			     format_linenum (numbuf0, p_input_line),
+			     format_linenum (numbuf1, p_hunk_beg));
 		    else if (! repl_could_be_missing)
-		      fatal ("duplicate `---' at line %ld; check line numbers at line %ld",
-			     p_input_line, p_hunk_beg + repl_beginning);
+		      fatal ("duplicate `---' at line %s; check line numbers at line %s",
+			     format_linenum (numbuf0, p_input_line),
+			     format_linenum (numbuf1,
+					     p_hunk_beg + repl_beginning));
 		    else
 		      {
 			repl_missing = TRUE;
@@ -928,11 +991,7 @@ another_hunk (difftype, rev)
 		p_Char[p_end] = '=';
 		for (s = buf;  *s && ! ISDIGIT (*s);  s++)
 		  continue;
-		if (!*s)
-		  malformed ();
-		p_newfirst = (LINENUM) atol (s);
-		while (ISDIGIT (*s))
-		  s++;
+		s = scan_linenum (s, &p_newfirst);
 		if (*s == ',')
 		  {
 		    do
@@ -941,7 +1000,8 @@ another_hunk (difftype, rev)
 			  malformed ();
 		      }
 		    while (! ISDIGIT (*s));
-		    p_repl_lines = (LINENUM) atol (s) - p_newfirst + 1;
+		    scan_linenum (s, &p_repl_lines);
+		    p_repl_lines += 1 - p_newfirst;
 		  }
 		else if (p_newfirst)
 		  p_repl_lines = 1;
@@ -1066,7 +1126,8 @@ another_hunk (difftype, rev)
 
     hunk_done:
 	if (p_end >=0 && !repl_beginning)
-	    fatal ("no `---' found in patch at line %ld", pch_hunk_beg ());
+	  fatal ("no `---' found in patch at line %s",
+		 format_linenum (numbuf0, pch_hunk_beg ()));
 
 	if (repl_missing) {
 
@@ -1132,8 +1193,10 @@ another_hunk (difftype, rev)
 		       && p_Char[fillsrc] != ' ')
 		    fillsrc++;
 		if (p_end < fillsrc || fillsrc == repl_beginning)
+		  {
 		    fatal ("replacement text or line numbers mangled in hunk at line %ld",
 			p_hunk_beg);
+		  }
 		p_line[filldst] = p_line[fillsrc];
 		p_Char[filldst] = p_Char[fillsrc];
 		p_len[filldst] = p_len[fillsrc];
@@ -1142,13 +1205,16 @@ another_hunk (difftype, rev)
 	    while (fillsrc <= p_end && fillsrc != repl_beginning)
 	      {
 		if (p_Char[fillsrc] == ' ')
-		  fatal ("replacement text or line numbers mangled in hunk at line %ld",
-			 p_hunk_beg);
+		  fatal ("replacement text or line numbers mangled in hunk at line %s",
+			 format_linenum (numbuf0, p_hunk_beg));
 		fillsrc++;
 	      }
 	    if (debug & 64)
-		printf("fillsrc %ld, filldst %ld, rb %ld, e+1 %ld\n",
-		    fillsrc,filldst,repl_beginning,p_end+1);
+	      printf ("fillsrc %s, filldst %s, rb %s, e+1 %s\n",
+		      format_linenum (numbuf0, fillsrc),
+		      format_linenum (numbuf1, filldst),
+		      format_linenum (numbuf2, repl_beginning),
+		      format_linenum (numbuf3, p_end + 1));
 	    assert(fillsrc==p_end+1 || fillsrc==repl_beginning);
 	    assert(filldst==p_end+1 || filldst==repl_beginning);
 	}
@@ -1167,29 +1233,18 @@ another_hunk (difftype, rev)
 	    next_intuit_at(line_beginning,p_input_line);
 	    return chars_read == (size_t) -1 ? -1 : 0;
 	}
-	s = buf+4;
-	if (!*s)
-	    malformed ();
-	p_first = (LINENUM) atol(s);
-	while (ISDIGIT (*s))
-	  s++;
-	if (*s == ',') {
-	    p_ptrn_lines = (LINENUM) atol(++s);
-	    while (ISDIGIT (*s))
-	      s++;
-	} else
+	s = scan_linenum (buf + 4, &p_first);
+	if (*s == ',')
+	    s = scan_linenum (s + 1, &p_ptrn_lines);
+	else
 	    p_ptrn_lines = 1;
 	if (*s == ' ') s++;
-	if (*s != '+' || !*++s)
+	if (*s != '+')
 	    malformed ();
-	p_newfirst = (LINENUM) atol(s);
-	while (ISDIGIT (*s))
-	  s++;
-	if (*s == ',') {
-	    p_repl_lines = (LINENUM) atol(++s);
-	    while (ISDIGIT (*s))
-	      s++;
-	} else
+	s = scan_linenum (s + 1, &p_newfirst);
+	if (*s == ',')
+	    s = scan_linenum (s + 1, &p_repl_lines);
+	else
 	    p_repl_lines = 1;
 	if (*s == ' ') s++;
 	if (*s != '@')
@@ -1205,14 +1260,18 @@ another_hunk (difftype, rev)
 	fillsrc = 1;
 	filldst = fillsrc + p_ptrn_lines;
 	p_end = filldst + p_repl_lines;
-	sprintf (buf,"*** %ld,%ld ****\n",p_first,p_first + p_ptrn_lines - 1);
+	sprintf (buf, "*** %s,%s ****\n",
+		 format_linenum (numbuf0, p_first),
+		 format_linenum (numbuf1, p_first + p_ptrn_lines - 1));
 	p_len[0] = strlen (buf);
 	if (! (p_line[0] = savestr (buf))) {
 	    p_end = -1;
 	    return -1;
 	}
 	p_Char[0] = '*';
-	sprintf (buf,"--- %ld,%ld ----\n",p_newfirst,p_newfirst+p_repl_lines-1);
+	sprintf (buf, "--- %s,%s ----\n",
+		 format_linenum (numbuf0, p_newfirst),
+		 format_linenum (numbuf1, p_newfirst + p_repl_lines - 1));
 	p_len[filldst] = strlen (buf);
 	if (! (p_line[filldst] = savestr (buf))) {
 	    p_end = 0;
@@ -1322,24 +1381,19 @@ another_hunk (difftype, rev)
 	    next_intuit_at(line_beginning,p_input_line);
 	    return chars_read == (size_t) -1 ? -1 : 0;
 	}
-	p_first = (LINENUM)atol(buf);
-	for (s = buf;  ISDIGIT (*s);  s++)
-	  continue;
+	s = scan_linenum (buf, &p_first);
 	if (*s == ',') {
-	    p_ptrn_lines = (LINENUM)atol(++s) - p_first + 1;
-	    while (ISDIGIT (*s))
-	      s++;
+	    s = scan_linenum (s + 1, &p_ptrn_lines);
+	    p_ptrn_lines += 1 - p_first;
 	}
 	else
 	    p_ptrn_lines = (*s != 'a');
 	hunk_type = *s;
 	if (hunk_type == 'a')
 	    p_first++;			/* do append rather than insert */
-	min = (LINENUM)atol(++s);
-	while (ISDIGIT (*s))
-	  s++;
+	s = scan_linenum (s + 1, &min);
 	if (*s == ',')
-	    max = (LINENUM)atol(++s);
+	    scan_linenum (s + 1, &max);
 	else
 	    max = min;
 	if (hunk_type == 'd')
@@ -1353,7 +1407,9 @@ another_hunk (difftype, rev)
 	    }
 	p_newfirst = min;
 	p_repl_lines = max - min + 1;
-	sprintf (buf, "*** %ld,%ld\n", p_first, p_first + p_ptrn_lines - 1);
+	sprintf (buf, "*** %s,%s\n",
+		 format_linenum (numbuf0, p_first),
+		 format_linenum (numbuf1, p_first + p_ptrn_lines - 1));
 	p_len[0] = strlen (buf);
 	if (! (p_line[0] = savestr (buf))) {
 	    p_end = -1;
@@ -1368,10 +1424,11 @@ another_hunk (difftype, rev)
 		return -1;
 	      }
 	    if (!chars_read)
-		fatal ("unexpected end of file in patch at line %ld",
-		  p_input_line);
+	      fatal ("unexpected end of file in patch at line %s",
+		     format_linenum (numbuf0, p_input_line));
 	    if (buf[0] != '<' || (buf[1] != ' ' && buf[1] != '\t'))
-		fatal ("`<' expected at line %ld of patch", p_input_line);
+	      fatal ("`<' expected at line %s of patch",
+		     format_linenum (numbuf0, p_input_line));
 	    chars_read -= 2 + (i == p_ptrn_lines && incomplete_line ());
 	    p_len[i] = chars_read;
 	    if (! (p_line[i] = savebuf (buf + 2, chars_read))) {
@@ -1388,12 +1445,15 @@ another_hunk (difftype, rev)
 		return -1;
 	      }
 	    if (! chars_read)
-		fatal ("unexpected end of file in patch at line %ld",
-		    p_input_line);
+	      fatal ("unexpected end of file in patch at line %s",
+		     format_linenum (numbuf0, p_input_line));
 	    if (*buf != '-')
-		fatal ("`---' expected at line %ld of patch", p_input_line);
+	      fatal ("`---' expected at line %s of patch",
+		     format_linenum (numbuf0, p_input_line));
 	}
-	sprintf (buf, "--- %ld,%ld\n", min, max);
+	sprintf (buf, "--- %s,%s\n",
+		 format_linenum (numbuf0, min),
+		 format_linenum (numbuf1, max));
 	p_len[i] = strlen (buf);
 	if (! (p_line[i] = savestr (buf))) {
 	    p_end = i-1;
@@ -1408,10 +1468,11 @@ another_hunk (difftype, rev)
 		return -1;
 	      }
 	    if (!chars_read)
-		fatal ("unexpected end of file in patch at line %ld",
-		    p_input_line);
+	      fatal ("unexpected end of file in patch at line %s",
+		     format_linenum (numbuf0, p_input_line));
 	    if (buf[0] != '>' || (buf[1] != ' ' && buf[1] != '\t'))
-		fatal ("`>' expected at line %ld of patch", p_input_line);
+	      fatal ("`>' expected at line %s of patch",
+		     format_linenum (numbuf0, p_input_line));
 	    chars_read -= 2 + (i == p_end && incomplete_line ());
 	    p_len[i] = chars_read;
 	    if (! (p_line[i] = savebuf (buf + 2, chars_read))) {
@@ -1433,7 +1494,8 @@ another_hunk (difftype, rev)
 		special = '^';
 	    else
 		special = ' ';
-	    fprintf (stderr, "%3ld %c %c ", i, p_Char[i], special);
+	    fprintf (stderr, "%s %c %c ", format_linenum (numbuf0, i),
+		     p_Char[i], special);
 	    pch_write_line (i, stderr);
 	    fflush (stderr);
 	}
@@ -1446,21 +1508,23 @@ another_hunk (difftype, rev)
 static size_t
 get_line ()
 {
-   return pget_line (p_indent, p_rfc934_nesting);
+   return pget_line (p_indent, p_rfc934_nesting, p_strip_trailing_cr);
 }
 
 /* Input a line from the patch file, worrying about indentation.
    Strip up to INDENT characters' worth of leading indentation.
    Then remove up to RFC934_NESTING instances of leading "- ".
+   If STRIP_TRAILING_CR is nonzero, remove any trailing carriage-return.
    Ignore any partial lines at end of input, but warn about them.
    Succeed if a line was read; it is terminated by "\n\0" for convenience.
    Return the number of characters read, including '\n' but not '\0'.
    Return -1 if we ran out of memory.  */
 
 static size_t
-pget_line (indent, rfc934_nesting)
+pget_line (indent, rfc934_nesting, strip_trailing_cr)
      int indent;
      int rfc934_nesting;
+     int strip_trailing_cr;
 {
   register FILE *fp = pfp;
   register int c;
@@ -1531,6 +1595,8 @@ pget_line (indent, rfc934_nesting)
 	goto patch_ends_in_middle_of_line;
     }
 
+  if (strip_trailing_cr && 2 <= i && b[i - 2] == '\r')
+    b[i-- - 2] = '\n';
   b[i] = '\0';
   p_input_line++;
   return i;
@@ -1803,14 +1869,16 @@ do_ed_script (ofp)
     register size_t chars_read;
 
     if (!skip_rest_of_patch) {
+	int exclusive = TMPOUTNAME_needs_removal ? 0 : O_EXCL;
 	assert (! inerrno);
-	copy_file (inname, TMPOUTNAME, instat.st_mode);
+	TMPOUTNAME_needs_removal = 1;
+	copy_file (inname, TMPOUTNAME, exclusive, instat.st_mode);
 	sprintf (buf, "%s %s%s", ed_program, verbosity == VERBOSE ? "" : "- ",
 		 TMPOUTNAME);
 	fflush (stdout);
 	pipefp = popen(buf, binary_transput ? "wb" : "w");
 	if (!pipefp)
-	  pfatal ("can't open pipe to `%s'", buf);
+	  pfatal ("Can't open pipe to %s", quotearg (buf));
     }
     for (;;) {
 	beginning_of_this_line = file_tell (pfp);
